@@ -1,6 +1,6 @@
 /*
     QMPlay2 is a video and audio player.
-    Copyright (C) 2010-2023  Błażej Szczygieł
+    Copyright (C) 2010-2024  Błażej Szczygieł
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU Lesser General Public License as published
@@ -25,6 +25,10 @@
 #include <Settings.hpp>
 #include <Packet.hpp>
 
+#ifdef Q_OS_ANDROID
+#   include <QFile>
+#endif
+
 #include <limits>
 
 extern "C"
@@ -36,6 +40,42 @@ extern "C"
     #include <libavutil/pixdesc.h>
 }
 
+#ifdef Q_OS_ANDROID
+static int readPacketQFile(void *opaque, uint8_t *buf, int bufSize)
+{
+    auto &f = *static_cast<QFile *>(opaque);
+    const int64_t bread =f.read(reinterpret_cast<char *>(buf), bufSize);
+    if (bread == 0)
+        return AVERROR_EOF;
+    return bread;
+}
+static int64_t seekQFile(void *opaque, int64_t offset, int whence)
+{
+    auto &f = *static_cast<QFile *>(opaque);
+    switch (whence)
+    {
+        case SEEK_SET:
+            if (!f.seek(offset))
+                return -1;
+            break;
+        case SEEK_CUR:
+            if (offset != 0)
+            {
+                if (!f.seek(f.pos() + offset))
+                    return -1;
+            }
+            break;
+        case SEEK_END:
+            if (!f.seek(f.size() - offset))
+                return -1;
+            break;
+        case AVSEEK_SIZE:
+            return f.size();
+    }
+    return f.pos();
+}
+#endif
+
 static void matroska_fix_ass_packet(AVRational stream_timebase, AVPacket *pkt)
 {
     AVBufferRef *line;
@@ -75,12 +115,12 @@ static void matroska_fix_ass_packet(AVRational stream_timebase, AVPacket *pkt)
     }
 }
 
-static QByteArray getTag(AVDictionary *metadata, const char *key, const bool deduplicate = true)
+static QByteArray getTag(AVDictionary *metadata, const char *key, const bool deduplicate = true, const int flags = 0)
 {
-    AVDictionaryEntry *avTag = av_dict_get(metadata, key, nullptr, 0);
+    AVDictionaryEntry *avTag = av_dict_get(metadata, key, nullptr, flags);
     if (avTag && avTag->value)
     {
-        const QByteArray tag = QByteArray(avTag->value);
+        const QByteArray tag = Functions::textWithFallbackEncoding(QByteArray(avTag->value));
         if (deduplicate)
         {
             // Workaround for duplicated tags separated by ';'.
@@ -196,14 +236,14 @@ private:
 
 /**/
 
-FormatContext::FormatContext(bool reconnectStreamed) :
+FormatContext::FormatContext(bool reconnectNetwork) :
     isError(false),
     currPos(0.0),
     abortCtx(new AbortContext),
     formatCtx(nullptr),
     packet(nullptr),
     oggHelper(nullptr),
-    reconnectStreamed(reconnectStreamed),
+    m_reconnectNetwork(reconnectNetwork),
     isPaused(false), fixMkvAss(false),
     isMetadataChanged(false),
     lastTime(0.0),
@@ -221,7 +261,7 @@ FormatContext::~FormatContext()
         av_packet_free(&packet);
     }
     delete oggHelper;
-    for (StreamInfo *streamInfo : qAsConst(streamsInfo))
+    for (StreamInfo *streamInfo : std::as_const(streamsInfo))
         delete streamInfo;
 }
 
@@ -328,14 +368,7 @@ QList<QMPlay2Tag> FormatContext::tags() const
     }
     if (isStreamed && !(value = getTag(formatCtx->metadata, "StreamTitle", false)).isEmpty())
     {
-        int idx = value.indexOf(" - ");
-        if (idx < 0)
-            tagList += {QString::number(QMPLAY2_TAG_TITLE), value};
-        else
-        {
-            tagList += {QString::number(QMPLAY2_TAG_TITLE), value.mid(idx + 3)};
-            tagList += {QString::number(QMPLAY2_TAG_ARTIST), value.mid(0, idx)};
-        }
+        tagList += {QString::number(QMPLAY2_TAG_TITLE), value};
     }
     else if (AVDictionary *dict = getMetadata())
     {
@@ -351,7 +384,7 @@ QList<QMPlay2Tag> FormatContext::tags() const
             tagList += {QString::number(QMPLAY2_TAG_DATE), value};
         if (!(value = getTag(dict, "comment")).isEmpty())
             tagList += {QString::number(QMPLAY2_TAG_COMMENT), value};
-        if (!(value = getTag(dict, "lyrics")).isEmpty())
+        if (!(value = getTag(dict, "lyrics", true, AV_DICT_IGNORE_SUFFIX)).isEmpty())
             tagList += {QString::number(QMPLAY2_TAG_LYRICS), value};
     }
     return tagList;
@@ -604,7 +637,6 @@ bool FormatContext::read(Packet &encoded, int &idx)
         }
         encoded.setDuration(newDuration);
     }
-    streamsTS[ff_idx] = encoded.ts();
 
     if (isStreamed)
     {
@@ -624,9 +656,11 @@ bool FormatContext::read(Packet &encoded, int &idx)
         return false;
     }
 
-    // Generate DTS for key frames if DTS doesn't exist (workaround for some M3U8 seekable streams)
+    // Generate DTS for key frames if DTS doesn't exist (e.g. workaround for some M3U8 seekable streams or other streams)
     if (encoded.hasKeyFrame() && !encoded.hasDts())
         encoded.setDts(nextDts.at(ff_idx));
+
+    streamsTS[ff_idx] = encoded.ts();
     nextDts[ff_idx] = encoded.ts() + encoded.duration();
 
     currPos = encoded.ts();
@@ -695,7 +729,15 @@ bool FormatContext::open(const QString &_url, const QString &param)
 
     AVInputFormat *inputFmt = nullptr;
     if (scheme == "file")
+    {
         isLocal = true;
+    }
+#ifdef Q_OS_ANDROID
+    else if (scheme == "content")
+    {
+        isLocal = true;
+    }
+#endif
     else
     {
         if (scheme != "rtsp")
@@ -712,14 +754,28 @@ bool FormatContext::open(const QString &_url, const QString &param)
     if (!inputFmt)
     {
         url = Functions::prepareFFmpegUrl(_url, options, false);
-        if (!isLocal && reconnectStreamed)
+        if (!isLocal && m_reconnectNetwork)
+        {
+            av_dict_set(&options, "reconnect", "1", 0);
             av_dict_set(&options, "reconnect_streamed", "1", 0);
+            av_dict_set(&options, "reconnect_on_network_error", "1", 0);
+            av_dict_set(&options, "reconnect_delay_max", "7", 0);
+        }
     }
 
     formatCtx = avformat_alloc_context();
     formatCtx->interrupt_callback.callback = (int(*)(void *))interruptCB;
     formatCtx->interrupt_callback.opaque = &abortCtx->isAborted;
 
+#ifdef Q_OS_ANDROID
+    if (isLocal && oggOffset < 0)
+    {
+        m_file = std::make_unique<QFile>(url);
+        m_file->open(QFile::ReadOnly);
+        formatCtx->pb = avio_alloc_context(nullptr, 0, false, m_file.get(), readPacketQFile, nullptr, seekQFile);
+    }
+#endif
+
     if (oggOffset >= 0)
     {
         oggHelper = new OggHelper(url, oggTrack, oggSize, formatCtx->interrupt_callback);
@@ -741,12 +797,27 @@ bool FormatContext::open(const QString &_url, const QString &param)
     if (!formatCtx || disabledDemuxers.contains(name()))
         return false;
 
+    auto isAvif = [&] {
+        if (formatCtx->nb_streams != 1 || !name().contains("mp4"))
+            return false;
+
+        AVStream *stream = formatCtx->streams[0];
+        if (stream->codecpar->codec_id != AV_CODEC_ID_AV1 || stream->nb_frames != 1)
+            return false;
+        if (stream->time_base.num != 1 || stream->time_base.den != 1)
+            return false;
+        if (stream->avg_frame_rate.num != 1 || stream->avg_frame_rate.den != 1)
+            return false;
+
+        return true;
+    };
+
     if (name() == "sdp")
     {
         isLocal = false;
         formatCtx->flags |= AVFMT_FLAG_NOBUFFER;
     }
-    else if (name().startsWith("image2") || name().endsWith("_pipe"))
+    else if (name().startsWith("image2") || name().endsWith("_pipe") || isAvif())
     {
         if (!settings.getBool("StillImages"))
             return false;
@@ -757,6 +828,12 @@ bool FormatContext::open(const QString &_url, const QString &param)
         formatCtx->flags |= AVFMT_FLAG_FAST_SEEK; //This should be set before "avformat_open_input", but seems to be working for MP3...
     }
 
+    if (name() == "mpegts")
+    {
+        // Workaround to detect initial audio parameters for some TS files
+        formatCtx->probesize *= 2;
+    }
+
     if (avformat_find_stream_info(formatCtx, nullptr) < 0)
         return false;
 
@@ -890,7 +967,7 @@ bool FormatContext::open(const QString &_url, const QString &param)
                 {
                     if (firstAudioIdx > -1)
                     {
-                        for (auto &&streamInfo : qAsConst(streamsInfo))
+                        for (auto &&streamInfo : std::as_const(streamsInfo))
                         {
                             if (streamInfo->params->codec_type == AVMEDIA_TYPE_AUDIO)
                                 streamInfo->is_default = false;
@@ -923,7 +1000,7 @@ void FormatContext::setStreamOffset(double offset)
 void FormatContext::selectStreams(const QSet<int> &selectedStreams)
 {
     m_allDiscarded = true;
-    for (AVStream *stream : qAsConst(streams))
+    for (AVStream *stream : std::as_const(streams))
     {
         if (stream->codecpar->codec_type == AVMEDIA_TYPE_DATA || stream->codecpar->codec_type == AVMEDIA_TYPE_ATTACHMENT)
         {
@@ -959,7 +1036,12 @@ StreamInfo *FormatContext::getStreamInfo(AVStream *stream) const
     if (const AVCodecDescriptor *codecDescr = avcodec_descriptor_get(streamInfo->params->codec_id))
     {
         if (codecDescr->props & AV_CODEC_PROP_TEXT_SUB)
-            streamInfo->must_decode = false;
+        {
+            if (codecDescr->id == AV_CODEC_ID_MOV_TEXT)
+                streamInfo->decode_to_ass = true;
+            else
+                streamInfo->must_decode = false;
+        }
 
         if (streamInfo->codec_name.isEmpty())
             streamInfo->codec_name = codecDescr->name;
